#!/usr/bin/env python
# coding=utf8
# File: test.py
import gym
import sys
import l3c.mazeworld
import time
import lmdb
import numpy
import argparse
from l3c.mazeworld import MazeTaskSampler
from l3c.mazeworld.agents import SmartSLAMAgent
from numpy import random

def test_agent_maze(n=15,
        maze_type="Discrete2D", 
        max_steps=1000, 
        task_type="NAVIGATION",
        density=0.40,
        n_landmarks=10,
        r_landmarks=0.40):
    print("\n\n--------\n\nRunning agents on maze_type = %s, task_type = %s, n = %d, steps = %s...\n\n"%(maze_type, task_type, n, max_steps))
    if(maze_type == "Discrete2D"):
        maze_env = gym.make("mazeworld-discrete-2D-v1", enable_render=False, max_steps=max_steps, task_type=task_type)
    elif(maze_type == "Discrete3D"):
        maze_env = gym.make("mazeworld-discrete-3D-v1", enable_render=False, max_steps=max_steps, task_type=task_type)
    elif(maze_type == "Continuous3D"):
        maze_env = gym.make("mazeworld-continuous-3D-v1", enable_render=False, max_steps=max_steps, task_type=task_type)
    else:
        raise Exception("No such maze world type %s"%task_type)

    task = MazeTaskSampler(n=n, allow_loops=True, 
            wall_density=density,
            landmarks_number=n_landmarks,
            landmarks_avg_reward=r_landmarks,
            commands_sequence = 10000,
            verbose=False)

    maze_env.set_task(task)

    # Must intialize agent after reset
    agent = SmartSLAMAgent(maze_env=maze_env, render=False)

    done=False
    observation = maze_env.reset()
    sum_reward = 0
    reward = 0
    observation_list = [observation]
    action_list = []
    reward_list = []
    map_list = []
    interval = 0

    while not done:
        action = agent.step(observation, reward)
        action_list.append(action)
        observation, reward, done, info = maze_env.step(action)
        loc_map = maze_env.maze_core.get_loc_map(map_range=3)
        reward_list.append(reward)
        observation_list.append(observation)
        map_list.append(loc_map)
        sum_reward += reward

    print("Finish running, sum reward = %f, steps = %d\n"%(sum_reward, len(observation_list)-1))

    return (numpy.asarray(observation_list, dtype=numpy.uint8), 
        numpy.asarray(action_list, dtype=numpy.uint8), 
        numpy.asarray(reward_list, dtype=numpy.float32),
        numpy.asarray(map_list, dtype=numpy.uint8))

if __name__=="__main__":
    # Parse the arguments, should include the output file name
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", type=str, default="maze.data.lmdb", help="output file name")
    parser.add_argument("--task_type", type=str, default="NAVIGATION", help="task type")
    parser.add_argument("--maze_type", type=str, default="Discrete3D", help="maze type")
    parser.add_argument("--max_steps", type=int, default=1000, help="max steps")
    parser.add_argument("--n", type=int, default=15, help="n")
    parser.add_argument("--density", type=float, default=0.36, help="density")
    parser.add_argument("--n_landmarks", type=int, default=10, help="n_landmarks")
    args = parser.parse_args()

    observations, actions, rewards, maps = test_agent_maze(
            maze_type=args.maze_type, 
            max_steps=args.max_steps,
            n=args.n, 
            task_type=args.task_type, 
            density=args.density, 
            n_landmarks=args.n_landmarks)

    # Convert observations, actions, and rewards to lmdb format and save file
    # Open the lmdb environment
    env = lmdb.open(args.output, map_size=1e12, sync=False)
    with env.begin(write=True) as txn:
        txn.put(b"observations", observations.tobytes())
        txn.put(b"actions", actions.tobytes())
        txn.put(b"rewards", rewards.tobytes())
        txn.put(b"maps", maps.tobytes())
    env.close()



"""
Core File of Maze Env
"""
import os
import numpy
import pygame
import random
from pygame import font
from copy import deepcopy
from numpy import random as npyrnd
from numpy.linalg import norm
from l3c.mazeworld.envs.dynamics import PI, PI_2, PI_4, PI2d, vector_move_with_collision
from l3c.mazeworld.envs.ray_caster_utils import maze_view
from l3c.mazeworld.envs.maze_task import MAZE_TASK_MANAGER
from l3c.mazeworld.envs.maze_base import MazeBase
from l3c.mazeworld.envs.ray_caster_utils import landmarks_rgb, landmarks_rgb_arr, paint_agent_arrow

class MazeCoreDiscrete3D(MazeBase):
    #Read Configurations
    def __init__(
            self,
            visibility_3D=12.0, #agent vision range
            fol_angle = 0.6 * PI, #Field of View
            resolution_horizon = 320, #resolution in horizontal
            resolution_vertical = 320, #resolution in vertical
            max_steps = 5000,
            task_type = "SURVIVAL"
        ):
        super(MazeCoreDiscrete3D, self).__init__(
                visibility_3D = visibility_3D,
                fol_angle = fol_angle,
                resolution_horizon = resolution_horizon,
                resolution_vertical = resolution_vertical,
                task_type = task_type,
                max_steps = max_steps
                )
        self._agent_arrow_theta = PI / 18.0

    def reset(self):
        if(self.task_type == "SURVIVAL"):
            #add the life bar
            self._lifebar_start_x = 0.10 * self.resolution_vertical
            self._lifebar_start_y = 0.10 * self.resolution_vertical
            self._lifebar_l = 0.80 * self.resolution_vertical
            self._lifebar_w = 0.05 * self.resolution_horizon
        elif(self.task_type == "NAVIGATION"):
            #add the navigation guidance bar
            self._navbar_start_x = 0.25 * self.resolution_vertical
            self._navbar_start_y = 0.10 * self.resolution_vertical
            self._navbar_l = 0.50 * self.resolution_vertical
            self._navbar_w = 0.05 * self.resolution_horizon
        else:
            raise Exception("No such task type: %s" % self._task_type)
        self._agent_ori_choice = numpy.asarray([0.0, 0.5, 1.0, 1.5], dtype="float32") * PI
        self._agent_ori_index = 0
        self._agent_ori = self._agent_ori_choice[self._agent_ori_index]
        return super(MazeCoreDiscrete3D, self).reset()

    def move(self, step):
        tmp_grid = deepcopy(self._agent_grid)
        if(self._agent_ori_index == 0):
            tmp_grid[0] += step
        elif(self._agent_ori_index == 1):
            tmp_grid[1] += step
        elif(self._agent_ori_index == 2):
            tmp_grid[0] -= step
        elif(self._agent_ori_index == 3):
            tmp_grid[1] -= step
        else:
            raise ValueError("Unexpected agent ori index: %d"%self._agent_ori_index)
        if(tmp_grid[0] >=0 and tmp_grid[0] < self._n and
                tmp_grid[1] >= 0  and tmp_grid[1] < self._n and
                self._cell_walls[tmp_grid[0], tmp_grid[1]] == 0):
            self._agent_grid = tmp_grid
            self._agent_loc = self.get_cell_center(self._agent_grid)

    def turn(self, direction):
        self._agent_ori_index += direction
        self._agent_ori_index = self._agent_ori_index % len(self._agent_ori_choice)
        self._agent_ori = self._agent_ori_choice[self._agent_ori_index]

    def do_action(self, action):
        assert numpy.shape(action) == (2,)
        assert abs(action[0]) < 2 and abs(action[1]) < 2
        self.turn(action[0])
        self.move(action[1])
        reward, done = self.evaluation_rule()
        self.update_observation()
        return reward, done

    def render_init(self, view_size):
        super(MazeCoreDiscrete3D, self).render_init(view_size)
        self._pos_conversion = self._render_cell_size / self._cell_size

    def render_observation(self):
        # Paint Observation
        view_obs_surf = pygame.transform.scale(self._obs_surf, (self._view_size, self._view_size))
        self._screen.blit(view_obs_surf, (0, 0))

        # Paint God-view
        agent_pos = numpy.array(self._agent_loc) * self._pos_conversion
        paint_agent_arrow(self._screen, pygame.Color("gray"), (self._view_size, 0), (agent_pos[0], agent_pos[1]), self._agent_ori, 
                0.4 * self._pos_conversion, 0.5 * self._pos_conversion)

    def movement_control(self, keys):
        #Keyboard control cases
        if keys[pygame.K_LEFT]:
            return (-1, 0)
        if keys[pygame.K_RIGHT]:
            return (1, 0)
        if keys[pygame.K_UP]:
            return (0, 1)
        if keys[pygame.K_DOWN]:
            return (0, -1)
        if keys[pygame.K_SPACE]:
            return (0, 0)
        return None

    def update_observation(self):
        if(self.task_type == "SURVIVAL"):
            self._observation, cell_exposed = maze_view(self._agent_loc, self._agent_ori, self._agent_height, 
                    self._cell_walls, self._cell_active_landmarks, self._cell_texts, self._cell_size, MAZE_TASK_MANAGER.grounds,
                    MAZE_TASK_MANAGER.ceil, self._wall_height, 1.0, self.visibility_3D, 0.20, 
                    self.fol_angle, self.resolution_horizon, self.resolution_vertical, landmarks_rgb_arr)
            lifebar_l = self._life / self._max_life * self._lifebar_l
            start_x = int(self._lifebar_start_x)
            start_y = int(self._lifebar_start_y)
            end_x = int(self._lifebar_start_x + lifebar_l)
            end_y = int(self._lifebar_start_y + self._lifebar_w)
            self._observation[start_x:end_x, start_y:end_y, 0] = 255
            self._observation[start_x:end_x, start_y:end_y, 1] = 0
            self._observation[start_x:end_x, start_y:end_y, 2] = 0
        elif(self.task_type == "NAVIGATION"):
            self._observation, cell_exposed = maze_view(self._agent_loc, self._agent_ori, self._agent_height, 
                    self._cell_walls, self._cell_landmarks, self._cell_texts, self._cell_size, MAZE_TASK_MANAGER.grounds,
                    MAZE_TASK_MANAGER.ceil, self._wall_height, 1.0, self.visibility_3D, 0.20, 
                    self.fol_angle, self.resolution_horizon, self.resolution_vertical, landmarks_rgb_arr)
            start_x = int(self._navbar_start_x)
            start_y = int(self._navbar_start_y)
            end_x = int(self._navbar_start_x + self._navbar_l)
            end_y = int(self._navbar_start_y + self._navbar_w)
            self._observation[start_x:end_x, start_y:end_y] = landmarks_rgb[self._command]
        self._obs_surf = pygame.surfarray.make_surface(self._observation)
        self._cell_exposed = numpy.logical_or(self._cell_exposed, cell_exposed)

    def get_observation(self):
        return numpy.copy(self._observation)



import gym
import sys
import argparse
import time
import l3c.mazeworld
from l3c.mazeworld import MazeTaskSampler
from l3c.mazeworld.agents import SmartSLAMAgent

if __name__=='__main__':
    parser = argparse.ArgumentParser(description='Playing the maze world demo with your keyboard')
    parser.add_argument('--maze_type', type=str, choices=["Discrete2D", "Discrete3D", "Continuous3D"], default="Continuous3D")
    parser.add_argument('--scale', type=int, default=15)
    parser.add_argument('--task_type', type=str, choices=["SURVIVAL", "NAVIGATION"], default="NAVIGATION")
    parser.add_argument('--max_steps', type=int, default=1000000)
    parser.add_argument('--density', type=float, default=0.30, help="Density of the walls satisfying that all spaces are connected")
    parser.add_argument('--visibility_2D', type=int, default=1, help="Grids vision range, only valid in 2D mode")
    parser.add_argument('--visibility_3D', type=float, default=12, help="3D vision range, Only valid in 3D mode")
    parser.add_argument('--wall_height', type=float, default=3.2, help="Only valid in 3D mode")
    parser.add_argument('--cell_size', type=float, default=2.0, help="Only valid in 3D mode")
    parser.add_argument('--step_reward', type=float, default=-0.01, help="Default rewards per-step")
    parser.add_argument('--n_landmarks', type=int, default=5, help="Number of landmarks, valid for both SURVIVAL and NAVIGATION task")
    parser.add_argument('--r_landmarks', type=float, default=0.50, help="Average rewards of the landmarks, only valid in SURVIVAL task")
    parser.add_argument('--cd_landmarks', type=int, default=200, help="Refresh interval of landmarks")
    parser.add_argument('--save_replay', type=str, default=None, help="Save the replay trajectory in file")
    parser.add_argument('--verbose', type=bool, default=False)

    args = parser.parse_args()

    if(args.maze_type == "Discrete2D"):
        maze_env = gym.make("mazeworld-discrete-2D-v1", enable_render=False, max_steps=args.max_steps, visibility_2D=args.visibility_2D, task_type=args.task_type)
    elif(args.maze_type == "Discrete3D"):
        maze_env = gym.make("mazeworld-discrete-3D-v1", enable_render=False, max_steps=args.max_steps, visibility_3D=args.visibility_3D, task_type=args.task_type)
    elif(args.maze_type == "Continuous3D"):
        maze_env = gym.make("mazeworld-continuous-3D-v1", enable_render=False, max_steps=args.max_steps, visibility_3D=args.visibility_3D, task_type=args.task_type)
    else:
        raise Exception("No such maze type %s"%args.maze_type)

    task = MazeTaskSampler(n=args.scale, allow_loops=True, 
            wall_density=args.density,
            cell_size=args.cell_size,
            wall_height=args.wall_height,
            step_reward=args.step_reward,
            landmarks_number=args.n_landmarks,
            landmarks_refresh_interval=args.cd_landmarks,
            landmarks_avg_reward=args.r_landmarks,
            verbose=True)
    maze_env.set_task(task)

    agent = SmartSLAMAgent(maze_env=maze_env, render=True)

    observation = maze_env.reset()
    done=False
    sum_reward = 0
    reward = 0

    while not done:
        action = agent.step(observation, reward)
        observation, reward, done, _ = maze_env.step(action)
        sum_reward += reward
        if(args.verbose):
            print("Instant r = %.2f, Accumulate r = %.2f" % (reward, sum_reward))
        if(maze_env.key_done):
            break
    print("Episode is over! You got %.2f score."%sum_reward)

    if(args.save_replay is not None):
        maze_env.save_trajectory(args.save_replay)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gym.envs.registration import register
from l3c.metalm.metalmv1 import MetaLMv1
from l3c.metalm.metalmv2 import MetaLMv2

register(
    id='meta-lm-v1',
    entry_point='l3c.metalm:MetaLMv1',
    kwargs={"V": 64,
        "n": 10,
        "l": 64,
        "e": 0.10,
        "L": 2048
    }
)

register(
    id='meta-lm-v2',
    entry_point='l3c.metalm:MetaLMv2',
    kwargs={"V": 64,
        "n": 16,
        "N": 16,
        "e": 0.10,
        "L": 2048
    }
)



#!/usr/bin/env python
# coding=utf8
# File: test.py
import gym
import sys
import l3c.mazeworld
import time
from l3c.mazeworld import MazeTaskSampler
from l3c.mazeworld.agents import SmartSLAMAgent
from numpy import random

def test_agent_maze(n=15,
        maze_type="Discrete2D", 
        max_steps=1000, 
        task_type="NAVIGATION",
        density=0.40,
        n_landmarks=10,
        r_landmarks=0.40):
    print("\n\n--------\n\nStart test for maze_type = %s, task_type = %s, n = %d, steps = %s...\n\n"%(maze_type, task_type, n, max_steps))
    if(maze_type == "Discrete2D"):
        maze_env = gym.make("mazeworld-discrete-2D-v1", enable_render=False, max_steps=max_steps, task_type=task_type)
    elif(maze_type == "Discrete3D"):
        maze_env = gym.make("mazeworld-discrete-3D-v1", enable_render=False, max_steps=max_steps, task_type=task_type)
    elif(maze_type == "Continuous3D"):
        maze_env = gym.make("mazeworld-continuous-3D-v1", enable_render=False, max_steps=max_steps, task_type=task_type)
    else:
        raise Exception("No such maze world type %s"%task_type)

    task = MazeTaskSampler(n=n, allow_loops=True, 
            wall_density=density,
            landmarks_number=n_landmarks,
            landmarks_avg_reward=r_landmarks,
            verbose=False)

    maze_env.set_task(task)

    # Must intialize agent after reset
    agent = SmartSLAMAgent(maze_env=maze_env, render=False)

    done=False
    observation = maze_env.reset()
    sum_reward = 0
    reward = 0
    while not done:
        action = agent.step(observation, reward)
        observation, reward, done, _ = maze_env.step(action)
        #print("action:", action, ", reward:", reward)
        time.sleep(0.01)
        sum_reward += reward
    print("...Test Finishes. Get score %f, for maze_type = %s task_type = %s, n = %d, steps = %s\n\n---------\n\n"%(sum_reward, maze_type, task_type, n, max_steps))

if __name__=="__main__":
    for n in [9, 15, 25]:
        for task_type in ["NAVIGATION", "SURVIVAL"]:
            for maze_type in ["Discrete2D", "Discrete3D", "Continuous3D"]:
                n_landmarks=random.randint(5,10)
                density=(random.random() + 0.50) * 0.30
                test_agent_maze(n=n, task_type=task_type, density=density, maze_type=maze_type, n_landmarks=n_landmarks)
    print("\n\nCongratulations!!!\n\nAll Tests Have Been Passed\n\n")



"""
Gym Environment For Maze3D
"""
import numpy
import gym
import pygame

from gym import error, spaces, utils
from gym.utils import seeding
from l3c.mazeworld.envs.maze_discrete_2d import MazeCoreDiscrete2D
from l3c.mazeworld.envs.maze_continuous_3d import MazeCoreContinuous3D
from l3c.mazeworld.envs.maze_discrete_3d import MazeCoreDiscrete3D

DISCRETE_ACTIONS=[(0, 0), (-1, 0), (1, 0), (0, -1), (0, 1)]

class MazeWorldDiscrete3D(gym.Env):
    def __init__(self, 
            enable_render=True,
            render_scale=480,
            resolution=(320, 320),
            max_steps=5000,
            visibility_3D=12.0,
            task_type="NAVIGATION",
            ):
        self.maze_type = "Discrete3D"
        self.enable_render = enable_render
        self.render_viewsize = render_scale
        self.maze_core = MazeCoreDiscrete3D(
                resolution_horizon = resolution[0],
                resolution_vertical = resolution[1],
                max_steps = max_steps,
                visibility_3D=visibility_3D,
                task_type = task_type,
                )

        # Turning Left/Right and go backward / forward
        self.action_space = spaces.Discrete(5)
        # observation is the x, y coordinate of the grid
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(resolution[0], resolution[1], 3), dtype=numpy.float32), 
                high=numpy.full((resolution[0], resolution[1], 3), 255, dtype=numpy.float32),
                dtype=numpy.float32)

        self.need_reset = True
        self.need_set_task = True

    def set_task(self, task_config):
        self.maze_core.set_task(task_config)
        self.need_set_task = False

    def reset(self):
        if(self.need_set_task):
            raise Exception("Must call \"set_task\" before reset")
        state = self.maze_core.reset()
        if(self.enable_render):
            self.maze_core.render_init(self.render_viewsize)
            self.keyboard_press = pygame.key.get_pressed()
        self.need_reset = False
        self.key_done = False
        return state

    def step(self, action=None):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")

        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(100) # 10 FPS
            action = self.maze_core.movement_control(self.keyboard_press)
        else:
            action = DISCRETE_ACTIONS[action]
            
        # In keyboard control, process only continues when key is pressed
        info = {"steps": self.maze_core.steps}
        if(action is None):
            return self.maze_core.get_observation(), 0, False, info 
        reward, done = self.maze_core.do_action(action)

        if(done):
            self.need_reset=True

        return self.maze_core.get_observation(), reward, done, info

    def render(self, mode="human"):
        if(mode != "human"):
            raise NotImplementedError("Only human mode is supported")
        if(self.enable_render):
            self.key_done, self.keyboard_press = self.maze_core.render_update()

    def get_loc_map(self, map_range=2):
        return self.maze_core.get_loc_map(map_rang=map_range)

    def save_trajectory(self, file_name):
        self.maze_core.render_trajectory(file_name)

class MazeWorldContinuous3D(gym.Env):
    def __init__(self, 
            enable_render=True,
            render_scale=480,
            resolution=(320, 320),
            max_steps = 5000,
            visibility_3D=12.0,
            task_type = "NAVIGATION"
            ):
        self.maze_type = "Continuous3D"
        self.enable_render = enable_render
        self.render_viewsize = render_scale
        self.maze_core = MazeCoreContinuous3D(
                resolution_horizon = resolution[0],
                resolution_vertical = resolution[1],
                max_steps = max_steps,
                visibility_3D=visibility_3D,
                task_type = task_type
                )

        # Turning Left/Right and go backward / forward
        self.action_space = spaces.Box(low=numpy.array([-1.0, -1.0]), 
                high=numpy.array([1.0, 1.0]), dtype=numpy.float32)
        # observation is the x, y coordinate of the grid
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(resolution[0], resolution[1], 3), dtype=numpy.float32), 
                high=numpy.full((resolution[0], resolution[1], 3), 256, dtype=numpy.float32),
                dtype=numpy.float32)

        self.need_reset = True
        self.need_set_task = True

    def set_task(self, task_config):
        self.maze_core.set_task(task_config)
        self.need_set_task = False

    def reset(self):
        if(self.need_set_task):
            raise Exception("Must call \"set_task\" before reset")
        state = self.maze_core.reset()
        if(self.enable_render):
            self.maze_core.render_init(self.render_viewsize)
            self.keyboard_press = pygame.key.get_pressed()
        self.need_reset = False
        self.key_done = False
        return state

    def step(self, action=None):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")
        reward = self.maze_core._step_reward
        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(20) # 50 FPS
            tr, ws = self.maze_core.movement_control(self.keyboard_press)
        else:
            tr = action[0]
            ws = action[1]

        # In keyboard control, process only continues when key is pressed
        info = {"steps": self.maze_core.steps}
        if(tr is None or ws is None):
            return self.maze_core.get_observation(), 0, False, info 

        reward, done = self.maze_core.do_action(tr, ws)

        if(done):
            self.need_reset=True

        return self.maze_core.get_observation(), reward, done, info

    def render(self, mode="human"):
        if(mode != "human"):
            raise NotImplementedError("Only human mode is supported")
        if(self.enable_render):
            self.key_done, self.keyboard_press = self.maze_core.render_update()

    def get_loc_map(self, map_range=2):
        return self.maze_core.get_loc_map(map_rang=map_range)

    def save_trajectory(self, file_name):
        self.maze_core.render_trajectory(file_name)

class MazeWorldDiscrete2D(gym.Env):
    def __init__(self,
            enable_render=True,
            render_scale=480,
            max_steps = 5000,
            task_type = "NAVIGATION",
            visibility_2D = 1):

        self.maze_type = "Discrete2D"
        self.enable_render = enable_render
        self.maze_core = MazeCoreDiscrete2D(visibility_2D=visibility_2D, max_steps=max_steps, task_type=task_type)
        self.render_viewsize = render_scale

        # Go EAST/WEST/SOUTH/NORTH
        self.action_space = spaces.Discrete(4)
        # observation is the x, y coordinate of the grid
        n_w = 2 * visibility_2D + 1
        self.observation_space = spaces.Box(low=numpy.zeros(shape=(n_w, n_w, 3), dtype=numpy.float32), 
                high=numpy.full((n_w, n_w, 3), 255, dtype=numpy.float32),
                dtype=numpy.float32)

        self.need_reset = True
        self.need_set_task = True

    def set_task(self, task_config):
        self.maze_core.set_task(task_config) 
        self.need_set_task = False

    def reset(self):
        if(self.need_set_task):
            raise Exception("Must call \"set_task\" before reset")
        state = self.maze_core.reset()
        if(self.enable_render):
            self.maze_core.render_init(self.render_viewsize)
            self.keyboard_press = pygame.key.get_pressed()
        self.need_reset = False
        self.key_done = False
        return state

    def step(self, action=None):
        if(self.need_reset):
            raise Exception("Must \"reset\" before doing any actions")
        if(action is None): # Only when there is no action input can we use keyboard control
            pygame.time.delay(100) # 10 FPS
            action = self.maze_core.movement_control(self.keyboard_press)
        else:
            action = DISCRETE_ACTIONS[action]
            
        info = {"steps": self.maze_core.steps}
        # In keyboard control, process only continues when key is pressed
        if(action is None):
            return self.maze_core.get_observation(), 0, False, info 

        reward, done = self.maze_core.do_action(action)

        if(done):
            self.need_reset=True

        return self.maze_core.get_observation(), reward, done, info

    def render(self, mode="human"):
        if(mode != "human"):
            raise NotImplementedError("Only human mode is supported")
        if(self.enable_render):
            self.key_done, self.keyboard_press = self.maze_core.render_update()

    def get_loc_map(self, map_range=2):
        return self.maze_core.get_loc_map(map_rang=map_range)

    def save_trajectory(self, file_name, additional=None):
        self.maze_core.render_trajectory(file_name, additional=additional)



import gym
import numpy
from numpy import random
import l3c.bandits

class ThompsonSampling(object):
    def __init__(self, arms):
        self.K = arms
        self.reset()

    def reset(self):
        self.record = numpy.zeros((self.K, 2)) + 1

    def update(self, idx, r):
        self.record[idx, 0] += r
        self.record[idx, 1] += 1.0 - r

    def select(self):
        val = random.beta(self.record[:,0], self.record[:,1])
        return numpy.argmax(val)

if __name__=="__main__":
    K=10
    env = gym.make("bandits-v0", arms=K)
    policy = ThompsonSampling(K)
    for i in range(10):
        task = env.sample_task()
        env.set_task(task)
        env.reset()
        policy.reset()
        total_r = 0
        done = False
        while not done:
            idx = policy.select()
            _, r, done, _ = env.step(idx)
            policy.update(idx, r)
            total_r += r
        print("%d th try, Thompson Sampling get %f, expected upper bound %f"%(i, total_r, env.expected_upperbound()))



from l3c.mazeworld.envs.maze_env import MazeWorldContinuous3D
from l3c.mazeworld.envs.maze_env import MazeWorldDiscrete3D
from l3c.mazeworld.envs.maze_env import MazeWorldDiscrete2D
from l3c.mazeworld.envs.maze_task import MAZE_TASK_MANAGER, MazeTaskSampler



import numpy
import math
import pygame
from .agent_base import AgentBase
from queue import Queue
from pygame import font
from l3c.mazeworld.envs.dynamics import PI
from l3c.mazeworld.envs.ray_caster_utils import landmarks_color, landmarks_rgb, landmarks_rgb_arr, paint_agent_arrow

class SmartSLAMAgent(AgentBase):
    def render_init(self, view_size=(640, 640)):
        """
        Initialize a God View With Landmarks
        """
        font.init()
        self._font = font.SysFont("Arial", 18)

        #Initialize the agent drawing
        self._render_cell_size_x = view_size[0] / self._nx
        self._render_cell_size_y = view_size[1] / self._ny
        self._view_size = view_size
        self._window_size = (view_size[0] * 2, view_size[1])

        self._pos_conversion_x = self._render_cell_size_x / self.maze_env.maze_core._cell_size
        self._pos_conversion_y = self._render_cell_size_y / self.maze_env.maze_core._cell_size

        self._screen = pygame.Surface(self._window_size)
        self._screen = pygame.display.set_mode(self._window_size)
        pygame.display.set_caption("AgentRender")
        self._surf_god = pygame.Surface(view_size)
        self._surf_god.fill(pygame.Color("white"))

        for x in range(self._nx):
            for y in range(self._ny):
                if(self._god_info[x,y] < 0):
                    pygame.draw.rect(self._surf_god, pygame.Color("black"), (x * self._render_cell_size_x, y * self._render_cell_size_y,
                            self._render_cell_size_x, self._render_cell_size_y), width=0)

    def render_update(self, observation):
        # paint landmarks
        self._screen.blit(self._surf_god, (0, 0))
        for landmarks_id, (x,y) in enumerate(self._landmarks_coordinates):
            if(self._landmarks_cd[landmarks_id] < 1):
                pygame.draw.rect(self._screen, landmarks_color(landmarks_id), 
                        (x * self._render_cell_size_x, y * self._render_cell_size_y,
                        self._render_cell_size_x, self._render_cell_size_y), width=0)
        # paint masks (mists)
        for x in range(self._nx):
            for y in range(self._ny):
                if(self._mask_info[x,y] < 1):
                    pygame.draw.rect(self._screen, pygame.Color("grey"), (x * self._render_cell_size_x, y * self._render_cell_size_y,
                            self._render_cell_size_x, self._render_cell_size_y), width=0)

        # add texts
        if(self.task_type is "SURVIVAL"):
            txt_life = self._font.render("Life: %f"%self._life, 0, pygame.Color("green"))
            self._screen.blit(txt_life,(90, 10))

        # paint agents
        agent_pos = [self._agent_loc[0] * self._pos_conversion_x, self._agent_loc[1] * self._pos_conversion_y]
        paint_agent_arrow(self._screen, pygame.Color("gray"), (0, 0), (agent_pos[0], agent_pos[1]), 0.5 * PI * self._agent_ori, 
                0.4 * self._pos_conversion_x, 0.5 * self._pos_conversion_x)

        # paint target trajectory
        for i in range(len(self._path)-1):
            factor = i / len(self._path)
            p = self._path[i]
            n = self._path[i+1]
            p = [(p[0] + 0.5) * self._render_cell_size_x, (p[1] + 0.5) *  self._render_cell_size_y]
            n = [(n[0] + 0.5) * self._render_cell_size_x, (n[1] + 0.5) *  self._render_cell_size_y]
            pygame.draw.line(self._screen, pygame.Color(int(255 * factor), int(255 * (1 - factor)), 128, 255), p, n, width=1)

        # paint observation
        obs_surf = pygame.surfarray.make_surface(observation)
        obs_surf = pygame.transform.scale(obs_surf, self._view_size)
        self._screen.blit(obs_surf, (self._view_size[0], 0))

        # display
        pygame.display.update()
        done = False
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                done=True
        keys = pygame.key.get_pressed()

    def update_cost_map(self, r_exp=0.25):
        # Calculate Shortest Distance using A*
        # In survival mode, consider the loss brought by rewards
        self._cost_map = 1e+6 * numpy.ones_like(self._god_info)
        refresh_list = Queue()
        refresh_list.put((self._cur_grid[0], self._cur_grid[1]))
        self._cost_map[self._cur_grid[0], self._cur_grid[1]] = 0
        while not refresh_list.empty():
            o_x, o_y = refresh_list.get()
            for d_x, d_y in self.neighbors:
                n_x = o_x + d_x
                n_y = o_y + d_y
                if(n_x >= self._nx or n_x < 0 or n_y >= self._ny or n_y < 0):
                    continue
                c_type = self._god_info[n_x, n_y]
                m_type = self._mask_info[n_x, n_y]
                if(c_type < 0 and m_type > 0):
                    continue
                elif(m_type < 1):
                    cost = 10
                elif(c_type > 0 and self._mask_info[n_x, n_y] > 0): # There is landmarks
                    if(self.task_type == "NAVIGATION"):
                        cost = 1
                    elif(self.task_type == "SURVIVAL"):
                        # Consider the extra costs of known traps
                        cost = 1
                        if(self._landmarks_rewards[c_type - 1] < 0.0 and self._landmarks_cd[c_type - 1] < 1):
                            cost += self._landmarks_rewards[c_type - 1] / self._step_reward
                else:
                    cost = 1
                if(self._cost_map[n_x, n_y] > self._cost_map[o_x, o_y] + cost):
                    self._cost_map[n_x, n_y] = self._cost_map[o_x, o_y] + cost
                    refresh_list.put((n_x, n_y))

    def policy(self, observation, r):
        self.update_cost_map()
        if(self.task_type=="SURVIVAL"):
            return self.policy_survival(observation, r)
        elif(self.task_type=="NAVIGATION"):
            return self.policy_navigation(observation, r)

    def policy_survival(self, observation, r):
        path_greedy, cost = self.navigate_landmarks_survival(0.50)
        path = path_greedy
        if(path is None or cost > -0.01):
            path_exp = self.exploration()
            if(path_exp is not None):
                path = path_exp
            elif(path is None):
                path = self._cur_grid
        self._path = path
        return self.path_to_action(path)

    def policy_navigation(self, observation, r):
        path_greedy = self.navigate_landmarks_navigate(self._command)
        path = path_greedy
        if(path_greedy is None):
            path_exp = self.exploration()
            if(path_exp is not None):
                path = path_exp
            else:
                path = self._cur_grid
        self._path = path
        return self.path_to_action(path)

    def path_to_action(self, path):
        if(self.maze_type=="Continuous3D"):
            return self.path_to_action_cont3d(path)
        elif(self.maze_type=="Discrete3D"):
            return self.path_to_action_disc3d(path)
        elif(self.maze_type=="Discrete2D"):
            return self.path_to_action_disc2d(path)

    def path_to_action_disc2d(self, path):
        if(len(path) < 2):
            return 0#(0, 0)
        d_x = path[1][0] - path[0][0]
        d_y = path[1][1] - path[0][1]
        if(d_x == -1 and d_y == 0):
            return 1#(-1, 0)
        if(d_x == 1 and d_y == 0):
            return 2#(1, 0)
        if(d_x == 0 and d_y == -1):
            return 3#(-1, 0)
        if(d_x == 0 and d_y == 1):
            return 4#(-1, 0)

    def path_to_action_disc3d(self, path):
        if(len(path) < 2):
            return 0 #(0, 0)
        d_x = path[1][0] - path[0][0]
        d_y = path[1][1] - path[0][1]
        req_ori = 2.0 * math.atan2(d_y, d_x) / PI
        deta_ori = req_ori - self._agent_ori
        deta_ori = int(deta_ori) % 4 + deta_ori - int(deta_ori)
        if(deta_ori > 2):
            deta_ori -= 4
        if(numpy.abs(deta_ori) < 0.2):
            return 4 #(0, 1)
        elif(deta_ori < 0):
            return 1 #(-1, 0)
        else:
            return 2 #(1, 0)
        #elif(numpy.abs(deta_ori) > 1.8 and self._god_info[path[1]] >=0 and self._mask_info[path[1]]>0):
        #    return 3 #(0, -1)

    def path_to_action_cont3d(self, path):
        if(len(path) < 2):
            d_x = path[0][0] + 0.5 - self._cur_grid_float[0]
            d_y = path[0][1] + 0.5 - self._cur_grid_float[1]
            deta_s = numpy.sqrt(d_x ** 2 + d_y ** 2)
            if(deta_s < 0.20):
                return (0, 0)
        else:
            d_x = path[1][0] + 0.5 - self._cur_grid_float[0]
            d_y = path[1][1] + 0.5 - self._cur_grid_float[1]
            deta_s = numpy.sqrt(d_x ** 2 + d_y ** 2)
        req_ori = 2.0 * math.atan2(d_y, d_x) / PI
        deta_ori = req_ori - self._agent_ori
        deta_ori = int(deta_ori) % 4 + deta_ori - int(deta_ori)
        if(deta_ori > 2):
            deta_ori -= 4
        if(numpy.abs(deta_ori) < 0.50):
            spd = min(deta_s, 1.0)
        elif(numpy.abs(deta_ori) > 1.95):
            spd = - min(deta_s, 1.0)
        else:
            spd = 0.0
        if(deta_ori < 0):
            turn = - min(numpy.abs(deta_ori), 1.0)
        else:
            turn = min(numpy.abs(deta_ori), 1.0)
        return (turn, spd)

    def retrieve_path(self, cost_map, goal_idx):
        path = [goal_idx]
        cost = cost_map[goal_idx]
        sel_x, sel_y = goal_idx
        iteration = 0
        while sel_x != self._cur_grid[0] or sel_y != self._cur_grid[1] :
            iteration += 1
            min_cost = cost
            for d_x, d_y in self.neighbors:
                n_x = sel_x + d_x
                n_y = sel_y + d_y
                if(n_x < 0 or n_x > self._nx - 1 or n_y < 0 or n_y > self._ny - 1):
                    continue
                if(cost_map[n_x, n_y] > 1e+4):
                    continue
                if(cost_map[n_x, n_y] < min_cost):
                    min_cost = cost_map[n_x, n_y]
                    sel_x = n_x
                    sel_y = n_y
                    path.insert(0, (n_x, n_y))
            cost=cost_map[sel_x, sel_y]
        return path

    def exploration(self):
        utility = self._cost_map + 10000 * self._mask_info
        if(numpy.argmin(utility) >= 10000):
            return None 
        target_idx = numpy.unravel_index(numpy.argmin(utility), utility.shape)
        return self.retrieve_path(self._cost_map, target_idx)

    def navigate_landmarks_navigate(self, landmarks_id):
        idxes = numpy.argwhere(self._god_info == landmarks_id + 1)
        for idx in idxes:
            if(self._mask_info[idx[0], idx[1]] < 1):
                continue
            else:
                return self.retrieve_path(self._cost_map, tuple(idx))
        return None

    def navigate_landmarks_survival(self, r_exp):
        cost_map = numpy.copy(self._cost_map)
        for i,idx in enumerate(self._landmarks_coordinates):
            if(i not in self._landmarks_visit and self._mask_info[idx] > 0):
                cost_map[idx] += r_exp / self._step_reward
            elif(self._landmarks_rewards[i] > 0.0 and self._mask_info[idx] > 0):
                cost_map[idx] += self._landmarks_cd[i] + self._landmarks_rewards[i] / self._step_reward
        target_idx = numpy.unravel_index(numpy.argmin(cost_map), cost_map.shape)
        return self.retrieve_path(self._cost_map, target_idx), cost_map[target_idx]




import sys
import numpy
from queue import Queue
from copy import deepcopy
from l3c.mazeworld.envs.dynamics import PI
from l3c.mazeworld.envs.maze_env import MazeWorldDiscrete2D, MazeWorldDiscrete3D, MazeWorldContinuous3D

class AgentBase(object):
    """
    Base class for agents
    Use this as parent to create new rule based agents
    """
    def __init__(self, **kwargs):
        self.render = False
        for k in kwargs:
            self.__dict__[k] = kwargs[k]
        if("maze_env" not in kwargs):
            raise Exception("Must use maze_env as arguments")
        self.task_type = self.maze_env.maze_core.task_type
        self.maze_type = self.maze_env.maze_type

        # Initialize information
        self._god_info = 1 - self.maze_env.maze_core._cell_walls + self.maze_env.maze_core._cell_landmarks
        if(self.task_type == "SURVIVAL"):
            self._landmarks_rewards = self.maze_env.maze_core._landmarks_rewards
        self._landmarks_coordinates = self.maze_env.maze_core._landmarks_coordinates
        self._step_reward = self.maze_env.maze_core._step_reward
        self._nx, self._ny = self._god_info.shape
        self.neighbors = [(-1, 0), (1, 0), (0, 1), (0, -1)]
        self._landmarks_visit = dict()

        # Render
        if(self.render):
            self.render_init()

    def render_init(self):
        raise NotImplementedError()

    def update_common_info(self):
        if(self.task_type == "NAVIGATION"):
            self._command = self.maze_env.maze_core._command
        if(self.task_type == "SURVIVAL"):
            self._life = self.maze_env.maze_core._life
        self._mask_info = self.maze_env.maze_core._cell_exposed
        self._agent_ori = (2.0 * self.maze_env.maze_core._agent_ori / PI)
        self._agent_loc = self.maze_env.maze_core._agent_loc
        self._cur_grid = deepcopy(self.maze_env.maze_core._agent_grid)
        self._cur_grid_float = deepcopy(self.maze_env.maze_core.get_loc_grid_float(self.maze_env.maze_core._agent_loc))
        self._landmarks_cd = []
        for cd in self.maze_env.maze_core._landmarks_refresh_countdown:
            if(cd < self.maze_env.maze_core._landmarks_refresh_interval):
                self._landmarks_cd.append(cd)
            else:
                self._landmarks_cd.append(0)
        self._exposed_info = self._god_info + self._mask_info - 1
        lid = self._god_info[self._cur_grid[0], self._cur_grid[1]]
        if(lid > 0):
            self._landmarks_visit[lid - 1] = 0

    def policy(self, observation, r):
        raise NotImplementedError()

    def render_update(self, observation):
        raise NotImplementedError()

    def step(self, observation, r):
        self.update_common_info()
        action = self.policy(observation, r)
        if(self.render):
            self.render_update(observation)
        return action
